#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "lis.h"

int lis_initialize2(char* argv, int options[], LIS_SCALAR params[]);
int lis_initialize_s(char *buf, int options[], LIS_SCALAR params[]);
int lis_initialize_solver(char *argv, int options[]);
int lis_initialize_precon(char *argv, int options[]);
int lis_initialize_print(char *argv, int options[]);
int lis_initialize_scale(char *argv, int options[]);
int lis_initialize_f(char *path, int options[], LIS_SCALAR params[]);

char *lis_solver_atoi[]  = {"cg", "bicg", "cgs", "bicgstab", "bicgstabl", "gpbicg", "tfqmr", "orthomin", "gmres", "jacobi", "gs", "sor"};
char *lis_precon_atoi[]  = {"none", "jacobi", "ilu", "ssor", "hybrid", "is", "sainv", "saamg", "ric2s"};
char *lis_storage_atoi[] = {"csr", "ccs", "msr", "dia", "ell", "jds", "bsr", "bsc", "vbr", "coo", "dns"};
char *lis_print_atoi[]  = {"none", "mem", "out", "all"};
char *lis_scale_atoi[]  = {"none", "jacobi", "symm_diag"};

#define LIS_PRINT_LEN	4
#define LIS_SCALE_LEN	3

void lis_display(void)
{
	printf("-i<a>           Iterative Solver(default:2)\n");
	printf("                  1|cg        CG          \n");
	printf("                  2|bicg      BiCG        \n");
	printf("                  3|cgs       CGS         \n");
	printf("                  4|bicgstab  BiCGSTAB    \n");
	printf("                  5|bicgstabl BiCGSTAB(l) \n");
	printf("                  6|gpbicg    GPBiCG      \n");
	printf("                  7|orthomin  Orthomin(m) \n");
	printf("                  8|gmress    GMRES(m)    \n");
	printf("                  9|tfqmr     TFQMR       \n");
	printf("                 10|jacobi    Jacobi      \n");
	printf("                 11|gs        Gauss-Seidel\n");
	printf("                 12|sor       SOR         \n");
	printf("-p<a>           Preconditioner(default:0)\n");
	printf("                  0|none      None          \n");
	printf("                  1|jacobi    Jacobi        \n");
	printf("                  2|ilu       ILU(k)        \n");
	printf("                  3|ssor      SSOR          \n");
	printf("                  4|hybrid    Hybrid        \n");
	printf("                  5|is        I+S           \n");
	printf("                  6|sainv     SAINV         \n");
	printf("                  7|saamg     SA-AMG        \n");
	printf("-maxiter<a>     Maximum number of iterations(default:1000)\n");
	printf("-tol<a>         Convergence Criteria(default:1.0e-12)\n");
	printf("-omega<a>       Damping or relaxation parameter used for SOR (default:1.9)\n");
	printf("-ssor_w<a>      Damping or relaxation parameter used for SSOR (default:1.0)\n");
	printf("-ilu_fill<a>    The level of fill-in (k) for ilu(k)(default:0)\n");
	printf("-ilu_relax<a>   Relaxation parameter for ilu(k)(default:1.0)\n");
	printf("-hybrid_i<a>    Iterative solver for hybrid(default:sor)\n");
	printf("-is_alpha<a>    Accelerated parameter for I+S type(default:1.0)\n");
	printf("-is_m<a>        Bandwith parameter for I+S type(default:3)\n");
	printf("-sainv_drop<a>  Specifies drop tolerance for SAINV(default:0.1)\n");
/*	printf("-ric2s_tau<a>   Specifies drop tolerance for RIC2S(default:0.05)\n");
	printf("-ric2s_sigma<a> Specifies drop tolerance for RIC2S(default:2.0)\n");
	printf("-ric2s_gamma<a> Specifies drop tolerance for RIC2S(default:1.0)\n");
	*/
}

void lis_version(void)
{
	printf("Version %s\n",LIS_VERSION);
}

void lis_set_default(int options[], LIS_SCALAR params[])
{
	options[LIS_OPTIONS_SOLVER]  = LIS_SOLVER_BICG;
	options[LIS_OPTIONS_PRECON]  = LIS_PRECON_TYPE_NONE;
	options[LIS_OPTIONS_OUTPUT]  = LIS_FALSE;
	options[LIS_OPTIONS_MAXITER] = 1000;
	options[LIS_OPTIONS_RESTART] = 40;
	options[LIS_OPTIONS_ELL]     = 2;
	options[LIS_OPTIONS_SCALE]   = LIS_SCALE_NONE;
	options[LIS_OPTIONS_FILL]    = 0;
	options[LIS_OPTIONS_M]       = 3;
	options[LIS_OPTIONS_PSOLVER]  = LIS_SOLVER_SOR;
	options[LIS_OPTIONS_PMAXITER] = 25;
	options[LIS_OPTIONS_PRESTART] = 40;
	options[LIS_OPTIONS_PELL]     = 2;
	options[LIS_OPTIONS_ISLEVEL]  = 1;

	params[LIS_PARAMS_RESID]     = 1.0e-12;
	params[LIS_PARAMS_W]         = 1.9;
	params[LIS_PARAMS_SSOR_W]    = 1.0;
	params[LIS_PARAMS_RELAX]     = 1.0;
	params[LIS_PARAMS_DROP]      = 1.0;
	params[LIS_PARAMS_ALPHA]     = 1.0;
	params[LIS_PARAMS_TAU]       = 0.05;
	params[LIS_PARAMS_SIGMA]     = 2.0;
	params[LIS_PARAMS_GAMMA]     = 1.0;
	params[LIS_PARAMS_PRESID]    = 1.0e-3;
	params[LIS_PARAMS_PW]        = 1.5;
}

int lis_initialize(int argc, char* argv[], char *fname, int options[], LIS_SCALAR params[])
{
	int i,t;
	char *p;

#ifdef _MPI
	MPI_Init(&argc, &argv);
#endif
	lis_set_default(options,params);
	if( fname==NULL )
	{
		for(i=1;i<argc;i++)
		{
			lis_initialize2(argv[i],options,params);
		}
	}
	else
	{
		lis_initialize_f(fname,options,params);
	}

	return LIS_SUCCESS;
}

int lis_finalize(void)
{
#ifdef _MPI
	MPI_Finalize();
#endif

	return LIS_SUCCESS;
}

int lis_initialize2(char* argv, int options[], LIS_SCALAR params[])
{
	int i,t;
	char *p;

	if( strncmp(argv, "-maxiter", 8)==0 )
	{
		p = argv + 8;
		sscanf(p, "%d", &options[LIS_OPTIONS_MAXITER]);
	}
	else if( strncmp(argv, "-tol", 4)==0 )
	{
		p = argv + 4;
		sscanf(p, "%lg", &params[LIS_PARAMS_RESID]);
	}
	else if( strncmp(argv, "-print", 6)==0 )
	{
		p = argv + 6;
		lis_initialize_print(p,options);
	}
	else if( strncmp(argv, "-scale", 6)==0 )
	{
		p = argv + 6;
		lis_initialize_scale(p,options);
	}
	else if( strncmp(argv, "-ssor_w", 7)==0 )
	{
		p = argv + 7;
		sscanf(p, "%lg", &params[LIS_PARAMS_SSOR_W]);
	}
	else if( strncmp(argv, "-ilu_fill", 9)==0 )
	{
		p = argv + 9;
		sscanf(p, "%d", &options[LIS_OPTIONS_FILL]);
	}
	else if( strncmp(argv, "-ilu_relax", 10)==0 )
	{
		p = argv + 10;
		sscanf(p, "%lg", &params[LIS_PARAMS_RELAX]);
	}
	else if( strncmp(argv, "-is_alpha", 9)==0 )
	{
		p = argv + 9;
		sscanf(p, "%lg", &params[LIS_PARAMS_ALPHA]);
	}
	else if( strncmp(argv, "-is_level", 9)==0 )
	{
		p = argv + 9;
		sscanf(p, "%d", &options[LIS_OPTIONS_ISLEVEL]);
	}
	else if( strncmp(argv, "-is_m", 5)==0 )
	{
		p = argv + 5;
		sscanf(p, "%d", &options[LIS_OPTIONS_M]);
	}
	else if( strncmp(argv, "-hybrid_maxiter", 15)==0 )
	{
		p = argv + 15;
		sscanf(p, "%d", &options[LIS_OPTIONS_PMAXITER]);
	}
	else if( strncmp(argv, "-hybrid_ell", 11)==0 )
	{
		p = argv + 11;
		sscanf(p, "%d", &options[LIS_OPTIONS_PELL]);
	}
	else if( strncmp(argv, "-hybrid_restart", 15)==0 )
	{
		p = argv + 15;
		sscanf(p, "%d", &options[LIS_OPTIONS_PRESTART]);
	}
	else if( strncmp(argv, "-hybrid_tol", 11)==0 )
	{
		p = argv + 11;
		sscanf(p, "%lg", &params[LIS_PARAMS_PRESID]);
	}
	else if( strncmp(argv, "-hybrid_w", 9)==0 )
	{
		p = argv + 9;
		sscanf(p, "%lg", &params[LIS_PARAMS_PW]);
	}
	else if( strncmp(argv, "-hybrid_i", 9)==0 )
	{
		p = argv + 9;
		sscanf(p, "%d", &options[LIS_OPTIONS_PSOLVER]);
	}
	else if( strncmp(argv, "-sainv_drop", 11)==0 )
	{
		p = argv + 11;
		sscanf(p, "%lg", &params[LIS_PARAMS_DROP]);
	}
	else if( strncmp(argv, "-ric2s_tau", 10)==0 )
	{
		p = argv + 10;
		sscanf(p, "%lg", &params[LIS_PARAMS_TAU]);
	}
	else if( strncmp(argv, "-ric2s_sigma", 12)==0 )
	{
		p = argv + 12;
		sscanf(p, "%lg", &params[LIS_PARAMS_SIGMA]);
	}
	else if( strncmp(argv, "-ric2s_gamma", 12)==0 )
	{
		p = argv + 12;
		sscanf(p, "%lg", &params[LIS_PARAMS_GAMMA]);
	}
	else if( strncmp(argv, "-restart", 8)==0 )
	{
		p = argv + 8;
		sscanf(p, "%d", &options[LIS_OPTIONS_RESTART]);
	}
	else if( strncmp(argv, "-ell", 4)==0 )
	{
		p = argv + 4;
		sscanf(p, "%d", &options[LIS_OPTIONS_ELL]);
	}
	else if( strncmp(argv, "-omega", 6)==0 )
	{
		p = argv + 6;
		sscanf(p, "%lg", &params[LIS_PARAMS_W]);
	}
	else if( strncmp(argv, "-i", 2)==0 )
	{
		p = argv + 2;
		lis_initialize_solver(p,options);
	}
	else if( strncmp(argv, "-p", 2)==0 )
	{
		p = argv + 2;
		lis_initialize_precon(p,options);
	}
	else if( strncmp(argv, "-f", 2)==0 )
	{
		p = argv + 2;
		lis_initialize_f(p,options,params);
	}
	else if( strncmp(argv, "-h", 2)==0 )
	{
		lis_display();
		exit(0);
	}
	else if( strncmp(argv, "-ver", 4)==0 )
	{
		lis_version();
		exit(0);
	}
	return LIS_SUCCESS;
}

int lis_initialize_solver(char *argv, int options[])
{
	int  i;
	char *p;

	if( argv[0]>='0' && argv[0]<='9' )
	{
		sscanf(argv, "%d", &options[LIS_OPTIONS_SOLVER]);
	}
	else
	{
		p = argv;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		for(i=0;i<LIS_SOLVER_LEN;i++)
		{
			if( strcmp(argv,lis_solver_atoi[i])==0 )
			{
				options[LIS_OPTIONS_SOLVER] = i+1;
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_initialize_precon(char *argv, int options[])
{
	int  i;
	char *p;

	if( argv[0]>='0' && argv[0]<='9' )
	{
		sscanf(argv, "%d", &options[LIS_OPTIONS_PRECON]);
	}
	else
	{
		p = argv;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		for(i=0;i<LIS_PRECON_TYPE_LEN;i++)
		{
			if( strcmp(argv,lis_precon_atoi[i])==0 )
			{
				options[LIS_OPTIONS_PRECON] = i;
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_initialize_print(char *argv, int options[])
{
	int  i;
	char *p;

	if( argv[0]>='0' && argv[0]<='3' )
	{
		sscanf(argv, "%d", &options[LIS_OPTIONS_OUTPUT]);
	}
	else
	{
		p = argv;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		for(i=0;i<LIS_PRINT_LEN;i++)
		{
			if( strcmp(argv,lis_print_atoi[i])==0 )
			{
				options[LIS_OPTIONS_OUTPUT] = i;
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_initialize_scale(char *argv, int options[])
{
	int  i;
	char *p;

	if( argv[0]>='0' && argv[0]<='2' )
	{
		sscanf(argv, "%d", &options[LIS_OPTIONS_SCALE]);
	}
	else
	{
		p = argv;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		for(i=0;i<LIS_SCALE_LEN;i++)
		{
			if( strcmp(argv,lis_scale_atoi[i])==0 )
			{
				options[LIS_OPTIONS_SCALE] = i;
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_initialize_f(char *path, int options[], LIS_SCALAR params[])
{
	int i;
	FILE *file;
	char str[1024];
	char seps[] = " \t\n\r";
	char *p;


	file = fopen(path, "r");
	if( file==NULL )
	{
		printf("cannot open %s\n", path);
		return LIS_FAILS;
	}
	while( fgets(str,sizeof(str),file)!=NULL )
	{
		p = str;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		p   = strtok( str, seps );

		while( p!=NULL )
		{
			lis_initialize2(p,options,params);
			p = strtok(NULL,seps);
		}
	}
	fclose(file);
	return LIS_SUCCESS;
}

int lis_initialize_s(char *buf, int options[], LIS_SCALAR params[])
{
	int i;
	char seps[] = " \t\n\r";
	char *p;

	lis_set_default(options,params);

	p = buf;
	while( *p!='\0' ) *p++ = (char)tolower(*p);
	p   = strtok( buf, seps );

	while( p!=NULL )
	{
		lis_initialize2(p,options,params);
		p = strtok(NULL,seps);
	}
	return LIS_SUCCESS;
}

int lis_set_options(char *text, int options[], LIS_SCALAR params[])
{
	int i;
	char seps[] = " \t\n\r";
	char *p;


	p = text;
	while( *p!='\0' ) *p++ = (char)tolower(*p);
	p   = strtok( text, seps );

	while( p!=NULL )
	{
		lis_initialize2(p,options,params);
		p = strtok(NULL,seps);
	}
	return LIS_SUCCESS;
}
