#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#ifdef _MPI
	#include <mpi.h>
#endif
#include "lis.h"
#include "lis_matrix.h"

void lis_commtable_destroy(LIS_COMMTABLE *table)
{
	if( table )
	{
		if( table->index_export ) lis_free( table->index_export );
		if( table->index_import ) lis_free( table->index_import );
		if( table->node_export ) lis_free( table->node_export );
		if( table->node_import ) lis_free( table->node_import );
#ifdef _MPI
		if( table->req1 ) lis_free( table->req1 );
		if( table->req2 ) lis_free( table->req2 );
		if( table->sta1 ) lis_free( table->sta1 );
		if( table->sta2 ) lis_free( table->sta2 );
#endif
		if( table->ws ) lis_free( table->ws );
		if( table->wr ) lis_free( table->wr );
		lis_free(table);
	}
}

#ifdef _MPI
int lis_commtable_duplicate(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
	int i;
	int n,np,nprocs,neibpetot,imnnz,exnnz;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *neibpe;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;

	LIS_COMMTABLE	*tin,*tout;

	tin       = Ain->commtable;
	n         = Ain->n;
	np        = Ain->np;
	nprocs    = Ain->nprocs;
	neibpetot = tin->neibpetot;
	imnnz     = tin->imnnz;
	exnnz     = tin->exnnz;
	comm      = tin->comm;

	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	req1         = NULL;
	sta1         = NULL;
	req2         = NULL;
	sta2         = NULL;
	tout         = NULL;
	err          = LIS_TRUE;
	do
	{
		tout = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( tout==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		import_node  = (int *)lis_malloc(imnnz*sizeof(int));
		if( import_node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		export_node  = (int *)lis_malloc(exnnz*sizeof(int));
		if( export_node==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( tout ) lis_free(tout);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	for(i=0;i<neibpetot;i++)   neibpe[i]       = tin->neibpe[i];
	for(i=0;i<neibpetot+1;i++) export_index[i] = tin->index_export[i];
	for(i=0;i<neibpetot+1;i++) import_index[i] = tin->index_import[i];
	for(i=0;i<exnnz;i++)       export_node[i]  = tin->node_export[i];
	for(i=0;i<imnnz;i++)       import_node[i]  = tin->node_import[i];

	tout->comm      = comm;
	tout->neibpetot = neibpetot;
	tout->imnnz     = imnnz;
	tout->exnnz     = exnnz;
	tout->wr        = wr;
	tout->ws        = ws;
	tout->index_export = export_index;
	tout->index_import = import_index;
	tout->node_export  = export_node;
	tout->node_import  = import_node;
	tout->neibpe       = neibpe;
	tout->req1         = req1;
	tout->req2         = req2;
	tout->sta1         = sta1;
	tout->sta2         = sta2;

	(*Aout)->commtable = tout;
	return LIS_SUCCESS;
}


int lis_matrix_g2l(LIS_MATRIX A)
{
	int err;
	switch( A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_matrix_g2l_crs(A);
		break;
	case LIS_MATRIX_CCS:
		err = lis_matrix_g2l_crs(A);
		break;
	case LIS_MATRIX_MSR:
		err = lis_matrix_g2l_msr(A);
		break;
	case LIS_MATRIX_DIA:
		err = lis_matrix_g2l_dia(A);
		break;
	case LIS_MATRIX_ELL:
		err = lis_matrix_g2l_ell(A);
		break;
	case LIS_MATRIX_JDS:
		err = lis_matrix_g2l_jds(A);
		break;
	case LIS_MATRIX_BSR:
		err = lis_matrix_g2l_bsr(A);
		break;
	case LIS_MATRIX_BSC:
		err = lis_matrix_g2l_bsc(A);
		break;
	case LIS_MATRIX_VBR:
		err = lis_matrix_g2l_vbr(A);
		break;
	case LIS_MATRIX_DNS:
		err = lis_matrix_g2l_dns(A);
		break;
	case LIS_MATRIX_COO:
		err = lis_matrix_g2l_coo(A);
		break;
	}
	if( err==LIS_NOT_IMPLEMENTED ) SETERR("not implemented\n");
	return err;
}

int lis_matrix_g2l_crs(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( nnz*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==0 )
				{
					np++;
					g2l_map[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = jj - is;
			}
			else
			{
				l_index[j] = g2l_map[jj];
			}
		}
		j = A->ptr[i];
//		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
	}
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
//				printf("jj=%d inpe=%d val=%f\n",jj,imneibpe[pe_map[jj]],A->value[j]);
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);

	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	/*
	req1         = NULL;
	sta1         = NULL;
	req2         = NULL;
	sta2         = NULL;
	*/
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j];
//				printf("jj=%d inpe=%d val=%f imnode=%d\n",jj,imneibpe[pe_map[jj]],A->value[j],A->index[j]);
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_msr(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,ndz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	ndz     = A->ndz;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->index[j];
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==0 )
				{
					np++;
					g2l_map[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->index[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = jj - is;
			}
			else
			{
				l_index[j] = g2l_map[jj];
			}
		}
		j = A->index[i];
//		lis_sort_iid(j,A->index[i+1]-1,l_index,A->index,A->value);
	}
	for(i=0;i<n+1;i++) l_index[i] = A->index[i];
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	/*
	req1         = NULL;
	sta1         = NULL;
	req2         = NULL;
	sta2         = NULL;
	*/
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;

		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;

		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;



		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;


		/*
		req1 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		*/
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
//		if( dmap ) lis_free(dmap);
//		if( pe ) lis_free(pe);
//		if( map ) lis_free(map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);

		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

//	n1 = 0;
	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j];
//				n1++;
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_dia(LIS_MATRIX A)
{
	int i,j,jj,k,kk,imcount,excount;
	int n,nnz,nnd,gn,np,nprocs;
	int is,ie,js,je,ns;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int	*index,*iw;
	int err;
	LIS_SCALAR *ws,*wr;
	LIS_SCALAR *value;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	nnd     = A->nnd;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	index     = NULL;
	value     = NULL;
	iw        = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( gn*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(is,-jj);
		je = _min(ie,gn-jj);
		for(i=js;i<je;i++)
		{
			kk =  jj+i;
			if( A->value[j*n+i-is]!=0.0 )
			{
				if( kk<is || kk>=ie )
				{
					if( g2l_map[kk]==0 )
					{
						np++;
						g2l_map[kk] = 1;
					}
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
			else
			{
				g2l_map[jj] = -1;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}
	for(jj=0;jj<gn;jj++)
	{
		if( g2l_map[jj]>=0 )
		{
			l_index[g2l_map[jj]] = jj;
		}
	}

	iw = (int *)lis_malloc( (n+gn)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n+gn;i++) iw[i] = 0;
	nnd = 0;
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(is,-jj);
		je = _min(ie,gn-jj);
		for(i=js;i<je;i++)
		{
			if( A->value[j*n+i-is]!=0.0 )
			{
				iw[n+g2l_map[jj+i]-i+is] = 1;
			}
		}
	}
	for(i=0;i<(n+gn);i++)
	{
		if( iw[i] > 0 ) nnd++;
	}
	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	j = 0;
	for( i=0; i<n+gn && j<nnd; i++ )
	{
		if( iw[i]>0 )
		{
			index[j]   = i-n;
			iw[i]  = j*n;
			j++;
		}
	}
	memset(value,0,n*nnd*sizeof(LIS_SCALAR));
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(is,-jj);
		je = _min(ie,gn-jj);
		for(i=js;i<je;i++)
		{
			if( A->value[j*n+i-is]!=0.0 )
			{
				value[iw[n+g2l_map[jj+i]-i+is] + i-is] = A->value[j*n+i-is];
			}
		}
	}

	A->g_index  = l_index;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		for(i=js;i<je;i++)
		{
			kk = jj+i;
			if( A->value[j*n+i]!=0.0 && pe_map[kk]!=my_rank && pe_map[kk] >= 0 )
			{
				imneibpe[pe_map[kk]]++;
				pe_map[kk] = - pe_map[kk] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;

		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;

		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;



		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;

		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);

		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

//	n1 = 0;
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		for(i=js;i<je;i++)
		{
			kk =  jj+i;
			if( A->value[j*n+i]!=0.0 && pe_map[kk]!=my_rank && pe_map[kk] < 0 )
			{
				pe_map[kk] = - pe_map[kk] - 1;
				k = imneibpe[pe_map[kk]]++;
				node[k] = kk;
				import_node[k] = g2l_map[kk];
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(A->index);
	lis_free(A->value);
	A->index    = index;
	A->value    = value;
	A->nnd      = nnd;

	lis_free(iw);
	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_ell(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,maxnzr,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	maxnzr  = A->maxnzr;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( n*maxnzr*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=0.0 )
			{
				jj = A->index[j*n + i];
				if( jj<is || jj>=ie )
				{
					if( g2l_map[jj]==0 )
					{
						np++;
						g2l_map[jj] = 1;
					}
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = A->index[j*n + i];
			if( jj>=is && jj<ie )
			{
				l_index[j*n + i] = jj - is;
			}
			else
			{
				l_index[j*n + i] = g2l_map[jj];
			}
		}
	}
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = A->g_index[j*n + i];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	/*
	req1         = NULL;
	sta1         = NULL;
	req2         = NULL;
	sta2         = NULL;
	*/
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;

		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;

		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;



		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;


		/*
		req1 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		*/
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
//		if( dmap ) lis_free(dmap);
//		if( pe ) lis_free(pe);
//		if( map ) lis_free(map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);

		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = A->g_index[j*n + i];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j*n +i];
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_jds(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,maxnzr,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	maxnzr  = A->maxnzr;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( nnz*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<n;i++) A->row[i] -= is;
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==0 )
				{
					np++;
					g2l_map[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = jj - is;
			}
			else
			{
				l_index[j] = g2l_map[jj];
			}
		}
	}
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	/*
	req1         = NULL;
	sta1         = NULL;
	req2         = NULL;
	sta2         = NULL;
	*/
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;

		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;

		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;



		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;


		/*
		req1 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		*/
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
//		if( dmap ) lis_free(dmap);
//		if( pe ) lis_free(pe);
//		if( map ) lis_free(map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);

		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j];
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_bsr(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos;
	int n,nnz,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	bnr     = A->bnr;
	bnc     = A->bnc;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( bnnz*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->bindex[j];
			for(k=0;k<bnc;k++)
			{
				if( jj<is || jj>=ie )
				{
					if( g2l_map[jj]==0 )
					{
						np++;
						g2l_map[jj] = 1;
					}
				}
				jj++;
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->bindex[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = (jj - is)/bnc;
			}
			else
			{
				l_index[j] = g2l_map[jj]/bnc;
			}
		}
//		j = A->ptr[i];
//		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
	}
	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += bnc;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;
		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				for(k=0;k<bnc;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = bnc*A->bindex[j]+k;
				}
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_bsc(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos;
	int n,nnz,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	bnr     = A->bnr;
	bnc     = A->bnc;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( bnnz*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->bindex[j];
			for(k=0;k<bnr;k++)
			{
				if( jj<is || jj>=ie )
				{
					if( g2l_map[jj]==0 )
					{
						np++;
						g2l_map[jj] = 1;
					}
				}
				jj++;
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->bindex[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = (jj - is)/bnr;
			}
			else
			{
				l_index[j] = g2l_map[jj]/bnr;
			}
		}
//		j = A->ptr[i];
//		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
	}
	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += bnr;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;
		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				for(k=0;k<bnr;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = bnr*A->bindex[j]+k;
				}
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_vbr(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos,kk,bj;
	int n,nnz,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie,ks,ke;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index,*l_col;
	int *g2l_map,*p2bindex;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( bnnz*sizeof(int) );
		if( l_index==NULL ) break;
		l_col = (int *)lis_malloc( (nc+1)*sizeof(int) );
		if( l_col==NULL ) break;
		p2bindex   = (int *)lis_malloc( gn*sizeof(int) );
		if( p2bindex==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( l_col ) lis_free(l_col);
		if( p2bindex ) lis_free(p2bindex);
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	kk = 0;
	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			bj  = A->bindex[j];
			ks   = A->col[bj];
			ke   = A->col[bj+1];
			for(jj=ks;jj<ke;jj++)
			{
				if( jj<is || jj>=ie )
				{
					if( g2l_map[jj]==0 )
					{
						np++;
						if( is<=ks && ie<ke )
						{
							kk++;
							g2l_map[jj] = 2;
						}
						else
						{
							g2l_map[jj] = 1;
						}
					}
				}
			}
		}
	}
	k  = n + kk;
	kk = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
			else if( g2l_map[jj]==2 )
			{
				g2l_map[jj] = kk++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<nc;i++)
	{
		p2bindex[i] = g2l_map[A->col[i]];
		l_col[i+1] = A->col[i+1] - A->col[i];
	}
	lis_sort_ii(0,nc-1,p2bindex,&l_col[1]);
	l_col[0] = 0;
	for(i=0;i<nc;i++)
	{
		l_col[i+1] += l_col[i];
	}

	for(i=0;i<nc;i++)
	{
		for(j=l_col[i];j<l_col[i+1];j++)
		{
			p2bindex[j] = i;
		}
	}
	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj                    = A->col[A->bindex[j]];
			bj                    = p2bindex[g2l_map[jj]];
			l_index[j]            = bj;
		}
	}

	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	t           = A->col;
	A->col      = l_col;
	A->g_col    = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = A->g_col[A->g_index[j]];
			if(jj>=gn)
			{
				printf("my_rank=%d jj=%d\n",my_rank,jj);
				continue;
			}
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += A->g_col[A->g_index[j]+1] - A->g_col[A->g_index[j]];
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);

	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;
		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = A->g_col[A->g_index[j]];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				bnc        = A->g_col[A->g_index[j]+1] - A->g_col[A->g_index[j]];
				for(k=0;k<bnc;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = A->col[A->bindex[j]]+k;
				}
			}
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);

	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_coo(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( nnz*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<nnz;i++)
	{
		A->row[i] -= is;
	}
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<nnz;i++)
	{
		jj = A->col[i];
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==0 )
			{
				np++;
				g2l_map[jj] = 1;
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<nnz;i++)
	{
		jj = A->col[i];
		if( jj>=is && jj<ie )
		{
			l_index[i] = jj - is;
		}
		else
		{
			l_index[i] = g2l_map[jj];
		}
//		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
	}
	t           = A->col;
	A->col      = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(i=0;i<nnz;i++)
	{
		jj = A->g_index[i];
		if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
		{
			imneibpe[pe_map[jj]]++;
			pe_map[jj] = - pe_map[jj] - 1;
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);

		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(i=0;i<nnz;i++)
	{
		jj = A->g_index[i];
		if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
		{
			pe_map[jj] = - pe_map[jj] - 1;
			k = imneibpe[pe_map[jj]]++;
			node[k] = jj;
			import_node[k] = A->col[i];
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);

	return LIS_SUCCESS;
}

int lis_matrix_g2l_dns(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,maxnzr,gn,np,nprocs;
	int is,ie,size;
	int my_rank;
	int neibpetot,n1,n2;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr,*tmp,v;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE *commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	maxnzr  = A->maxnzr;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = gn;

	l_index   = NULL;
	pe_map    = NULL;
	g2l_map   = NULL;
	imneibpe  = NULL;
	exneibpe  = NULL;
	req1      = NULL;
	sta1      = NULL;
	req2      = NULL;
	sta2      = NULL;
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( gn*sizeof(int) );
		if( l_index==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		pe_map = (int *)lis_malloc( gn*sizeof(int) );
		if( pe_map==NULL ) break;
		g2l_map = (int *)lis_malloc(gn*sizeof(int));
		if( g2l_map==NULL ) break;
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( imneibpe==NULL ) break;
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));
		if( exneibpe==NULL ) break;
		tmp = (LIS_SCALAR *)lis_malloc( n*n*sizeof(LIS_SCALAR) );
		if( tmp==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( pe_map ) lis_free(pe_map);
		if( g2l_map ) lis_free(g2l_map);
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( tmp ) lis_free(tmp);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			g2l_map[jj] = k;
			k++;
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}
	for(jj=0;jj<gn;jj++)
	{
		l_index[g2l_map[jj]] = jj;
	}

	memcpy(tmp,&A->value[is*n],n*n*sizeof(LIS_SCALAR));
	for(j=is-1;j>=0;j--)
	{
		memcpy(&A->value[(j+n)*n],&A->value[j*n],n*sizeof(LIS_SCALAR));
	}
	memcpy(&A->value[0],tmp,n*n*sizeof(LIS_SCALAR));
	A->g_index  = l_index;
	A->np       = gn;

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}


	for(jj=0;jj<gn;jj++)
	{
		if( pe_map[jj]!=my_rank )
		{
			imneibpe[pe_map[jj]]++;
		}
	}

	n1=0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);
			n1++;
		}
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);


	neibpetot = 0;
	imcount   = 0;
	excount   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpetot++;
			imcount += imneibpe[i];
			excount += exneibpe[i];
		}
	}


	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	commtable    = NULL;
	err          = LIS_TRUE;
	do
	{
		commtable = (LIS_COMMTABLE *)lis_malloc( sizeof(LIS_COMMTABLE) );
		if( commtable==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		import_node  = (int *)lis_malloc(imcount*sizeof(int));
		if( import_node==NULL ) break;
		node  = (int *)lis_malloc(imcount*sizeof(int));
		if( node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		export_node  = (int *)lis_malloc(excount*sizeof(int));
		if( export_node==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( imneibpe ) lis_free(imneibpe);
		if( exneibpe ) lis_free(exneibpe);
		if( tmp ) lis_free(tmp);

		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( node  ) lis_free(node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( commtable ) lis_free(commtable);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	n1              = 0;
	n2              = 0;
	import_index[0] = 0;
	export_index[0] = 0;
	for(i=0;i<nprocs && n1<neibpetot;i++)
	{
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )
		{
			neibpe[n1]         = i;
			if( imneibpe[i]!=0 )
			{
				import_index[n1+1] = import_index[n1] + imneibpe[i];
				imneibpe[i] = import_index[n1];
			}
			else
			{
				import_index[n1+1] = import_index[n1];
			}
			if( exneibpe[i]!=0 )
			{
				export_index[n1+1] = export_index[n1] + exneibpe[i];
			}
			else
			{
				export_index[n1+1] = export_index[n1];
			}
			n1++;
		}
	}

	for(j=0;j<gn;j++)
	{
		jj = A->g_index[j];
		if( pe_map[jj]!=my_rank )
		{
			k = imneibpe[pe_map[jj]]++;
			node[k] = jj;
			import_node[k] = j;
		}
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = import_index[i];
		n2 = import_index[i+1] - import_index[i];
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);
	}
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i];
		n2 = export_index[i+1] - export_index[i];
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);
	}
	MPI_Waitall(nprocs-1, req1, sta1);
	MPI_Waitall(nprocs-1, req2, sta2);
	for(i=0;i<neibpetot;i++)
	{
		k = neibpe[i];
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			export_node[j] = g2l_map[export_node[j]];
		}
	}

	commtable->comm         = comm;
	commtable->neibpetot    = neibpetot;
	commtable->imnnz        = imcount;
	commtable->exnnz        = excount;
	commtable->neibpe       = neibpe;
	commtable->index_export = export_index; 
	commtable->index_import = import_index; 
	commtable->node_export  = export_node; 
	commtable->node_import  = import_node; 
	commtable->req1         = req1;
	commtable->req2         = req2;
	commtable->sta1         = sta1;
	commtable->sta2         = sta2;
	commtable->ws           = ws;
	commtable->wr           = wr;

	A->commtable = commtable;

	lis_free(pe_map);
	lis_free(g2l_map);
	lis_free(imneibpe);
	lis_free(exneibpe);
	lis_free(node);
	lis_free(tmp);

	return LIS_SUCCESS;
}

#endif

#ifdef _MPI
int lis_send_recv(LIS_COMMTABLE *commtable, LIS_SCALAR x[])
{
	int neib,i,is,inum,neibpetot;
	LIS_SCALAR *ws,*wr;

	neibpetot = commtable->neibpetot;
	ws        = commtable->ws;
	wr        = commtable->wr;
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_export[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			ws[i] = x[commtable->node_export[i]];
		}
		MPI_Isend(&ws[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req1[neib]);
	}
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		MPI_Irecv(&wr[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req2[neib]);
	}
	MPI_Waitall(neibpetot, commtable->req2, commtable->sta2);
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			x[commtable->node_import[i]] = wr[i];
		}
	}
	MPI_Waitall(neibpetot, commtable->req1, commtable->sta1);
	return LIS_SUCCESS;
}
int lis_reduce(LIS_COMMTABLE *commtable, LIS_SCALAR x[])
{
	int neib,i,is,inum,neibpetot;
	LIS_SCALAR *ws,*wr;

	neibpetot = commtable->neibpetot;
	ws        = commtable->ws;
	wr        = commtable->wr;
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			ws[i] = x[commtable->node_import[i]];
		}
		MPI_Isend(&ws[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req1[neib]);
	}
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_export[neib+1] - is;
		MPI_Irecv(&wr[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req2[neib]);
	}
	MPI_Waitall(neibpetot, commtable->req2, commtable->sta2);
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_import[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			x[commtable->node_export[i]] += wr[i];
		}
	}
	MPI_Waitall(neibpetot, commtable->req1, commtable->sta1);
	return LIS_SUCCESS;
}
#endif


