PROGRAM solver_test
  USE count_time_mod
  USE SOLVER_AMGCG
  IMPLICIT NONE
  INCLUDE 'precision.inc'
  INCLUDE 'mpif.h'

  REAL(kind=KREAL), DIMENSION (:), ALLOCATABLE :: DELPHI, DELPHI_o

  !C-- initial matrix form
  REAL   (kind=KREAL), POINTER :: DIAG(:)
  REAL   (kind=KREAL), POINTER :: BFORCE(:)
  INTEGER(kind=kint ), POINTER :: ISTACK_U(:), ISTACK_L(:)
  INTEGER(kind=kint ), POINTER ::   ITEM_U(:),   ITEM_L(:)
  REAL   (kind=KREAL), POINTER :: AICCG_U(:)
  REAL   (kind=KREAL), POINTER :: AICCG_L(:)

  !C-- redistributed matrix form
  REAL   (kind=KREAL), POINTER :: D(:), B(:)
  INTEGER(kind=kint ), POINTER :: INU(:), INL(:)
  INTEGER(kind=kint ), POINTER :: IAU(:), IAL(:)
  REAL   (kind=KREAL), POINTER :: AU(:)
  REAL   (kind=KREAL), POINTER :: AL(:)
  INTEGER(kind=kint ), POINTER :: NOD_IMPORT(:)
  INTEGER(kind=kint ), POINTER :: NOD_EXPORT(:)
  INTEGER(kind=kint ), POINTER :: NEIBPE(:)
  INTEGER(kind=kint ), POINTER :: STACK_IMPORT(:)
  INTEGER(kind=kint ), POINTER :: STACK_EXPORT(:)

  !C-- reredistributed matrix form
  REAL   (kind=KREAL), POINTER :: D_o(:), B_o(:)
  INTEGER(kind=kint ), POINTER :: INU_o(:), INL_o(:)
  INTEGER(kind=kint ), POINTER :: IAU_o(:), IAL_o(:)
  REAL   (kind=KREAL), POINTER :: AU_o(:)
  REAL   (kind=KREAL), POINTER :: AL_o(:)
  INTEGER(kind=kint ), POINTER :: NOD_IMPORT_o(:)
  INTEGER(kind=kint ), POINTER :: NOD_EXPORT_o(:)
  INTEGER(kind=kint ), POINTER :: NEIBPE_o(:)
  INTEGER(kind=kint ), POINTER :: STACK_IMPORT_o(:)
  INTEGER(kind=kint ), POINTER :: STACK_EXPORT_o(:)
  INTEGER(kind=kint ) :: N_o,NP_o,NEIBPETOT_o

  REAL   (kind=kreal) :: REALARRAY(10)
  INTEGER(kind=kint ) ::  INTARRAY(10)
  INTEGER             ::  PRESET, ITERactual

  INTEGER(kind=kint ) :: PETOT, NEIBPETOT
  INTEGER(kind=kint ) :: my_rank

  INTEGER(kind=kint ) :: npsi, npse, npl, npu, nb, n, np, ierr, intnodtot, nodtot
  REAL(kind=kreal)    :: STARTTIME, ENDTIME, RTIME

  CHARACTER(len= 4 )  :: penum, penum_left, penumtot_char, petot_left
  CHARACTER(len=10 )  :: fname
  CHARACTER(len=80)   :: MATfile(128)
  INTEGER(kind=kint ) :: i, j, k

  INTEGER(kind=kint ) :: SOLVER_COMM, ERROR, ITERactural
  CHARACTER(len=20)   :: METHOD, PRECOND, DIST
  LOGICAL :: dist_f, open_f

  !C-- distribute file access      
  INTEGER(kind=kint ) :: CRITICAL_PROC, MY_SCORE, TOKEN

  !C-- communication
  REAL(kind=kreal), ALLOCATABLE :: WS(:), WR(:)

  !C--
  INTEGER(kind=kint), POINTER :: order(:), nds_index(:)
  INTEGER(kind=kint), POINTER :: order_o(:)

  !C--
  INTEGER(kind=kint) :: nparts
  REAL(kind=kfloat), ALLOCATABLE :: tpwgts(:)

  REAL(kind=kreal) :: RESID
  integer(kind=kint) :: iter


  !C
  !C-- init. MPI
  CALL MPI_INIT      (ierr)
  CALL MPI_COMM_SIZE (MPI_COMM_WORLD, PETOT, ierr )
  CALL MPI_COMM_RANK (MPI_COMM_WORLD, my_rank, ierr )
  CALL MPI_COMM_DUP  (MPI_COMM_WORLD, SOLVER_COMM, ierr)
  
  !C
  !C-- CNTL DATA
  WRITE(penumtot_char,'(i4)')  PETOT
  petot_left = ADJUSTL(penumtot_char)
  j=LEN_TRIM(petot_left)
  DO i=1,PETOT
     WRITE(penum,'(i4)')  i-1
     penum_left = ADJUSTL(penum)
     WRITE(fname,'(a6,a4)')'matIN.',penum_left
!!$         Matfile(i)='/home/fujii/mg/work/problems/'//petot_left(1:j)//'PE/'//fname
     Matfile(i)='/tmp/'//fname
  END DO

  !C
  !C +-------------+
  !C | MATRIX file |
  !C +-------------+
  !C===

!!$      TOKEN=PETOT
!!$      MY_SCORE = my_rank
!!$      CRITICAL_PROC = 0
!!$      DO while(CRITICAL_PROC /= TOKEN) 
!!$      CALL MPI_ALLREDUCE(MY_SCORE,CRITICAL_PROC,1,MPI_INTEGER,MPI_MIN,  &
!!$     & SOLVER_COMM,ierr)
!!$      IF(my_rank==CRITICAL_PROC) then
!!$      MY_SCORE=TOKEN
!!$      WRITE(*,*) my_rank,"reading.."
  !C############### critical section: start


  !C-- single process has matrix.
  IF(my_rank >0) THEN
     NODTOT = 0
  END IF

  open_f=.false.
!!$  if(my_rank == 0) then
  !C     open (15, file=MATfile(my_rank+1), status='unknown',form='unformatted')
  OPEN(15, file = MATfile(my_rank + 1), form = 'unformatted', status = 'old', err = 100)
!!$      read (15,'(10i8)') N, NP, NPL, NPU, NB, NEIBPETOT
  READ (15) N, NP, NPL, NPU, NB, NEIBPETOT
  open_f=.true.
  INTNODTOT = N
  NODTOT    = NP


  ALLOCATE (ISTACK_L(0:NP), ITEM_L(NPL))
  ALLOCATE (ISTACK_U(0:NP), ITEM_U(NPU))
  ALLOCATE (AICCG_L(NPL), AICCG_U(NPU), DIAG(NP))
  ALLOCATE (BFORCE (NP))

!!$      read (15,'(10i8)') (ISTACK_L(k), k= 1, NP)
!!$      read (15,'(10i8)') (ISTACK_U(k), k= 1, NP)
!!$
!!$      read (15,'(10i8)') (  ITEM_L(k), k= 1, NPL)
!!$      read (15,'(10i8)') (  ITEM_U(k), k= 1, NPU)
!!$
!!$      read (15,'(5e27.20)') (AICCG_L(k), k= 1,NPL)
!!$      read (15,'(5e27.20)') (AICCG_U(k), k= 1,NPU)
!!$      
!!$      read (15,'(5e27.20)') (DIAG  (i), i= 1,NP)
!!$      read (15,'(5e27.20)') (BFORCE(i), i= 1,NP)
!!$
  READ (15) (ISTACK_L(k), k= 1, NP)
  READ (15) (ISTACK_U(k), k= 1, NP)

  READ (15) (ITEM_L(k), k= 1, NPL)
  READ (15) (ITEM_U(k), k= 1, NPU)

  READ (15) (AICCG_L(k), k= 1,NPL)
  READ (15) (AICCG_U(k), k= 1,NPU)

  READ (15) (DIAG  (i), i= 1,NP)
  READ (15) (BFORCE(i), i= 1,NP)


  ISTACK_U(0)= 0
  ISTACK_L(0)= 0

  ALLOCATE (NEIBPE(NEIBPETOT))
  IF (PETOT.NE.1) THEN
!!$        read (15,'(10i8)') (NEIBPE(k), k= 1, NEIBPETOT)
     READ (15) (NEIBPE(k), k= 1, NEIBPETOT)
     ALLOCATE (STACK_IMPORT(0:NEIBPETOT))
     ALLOCATE (STACK_EXPORT(0:NEIBPETOT))

     STACK_IMPORT(0)= 0
     STACK_EXPORT(0)= 0

!!$        read (15,'(10i8)') (STACK_IMPORT(k), k= 1, NEIBPETOT)
!!$        read (15,'(10i8)') (STACK_EXPORT(k), k= 1, NEIBPETOT)
     READ (15) (STACK_IMPORT(k), k= 1, NEIBPETOT)
     READ (15) (STACK_EXPORT(k), k= 1, NEIBPETOT)

     NPSI= STACK_IMPORT(NEIBPETOT)
     NPSE= STACK_EXPORT(NEIBPETOT)

     ALLOCATE (NOD_IMPORT(NPSI))
     ALLOCATE (NOD_EXPORT(NPSE))
!!$        read (15,'(10i8)') (NOD_IMPORT(k), k= 1, NPSI)
!!$        read (15,'(10i8)') (NOD_EXPORT(k), k= 1, NPSE)

     READ (15) (NOD_IMPORT(k), k= 1, NPSI)
     READ (15) (NOD_EXPORT(k), k= 1, NPSE)
  ENDIF
  CLOSE (15)

  if(.not.open_f) then
100      write(*,*) "no problem file @ PE:",my_rank

       N = 0; NP = 0; NPL = 0; NPU = 0; NB = 1; NEIBPETOT = 0
       ALLOCATE (ISTACK_L(0:NP), ITEM_L(NPL))
       ALLOCATE (ISTACK_U(0:NP), ITEM_U(NPU))
       ALLOCATE (AICCG_L(NPL), AICCG_U(NPU), DIAG(NP))
       ALLOCATE (BFORCE (NP))
       ALLOCATE (NEIBPE(NEIBPETOT))
       IF (PETOT.NE.1) THEN
          ALLOCATE (STACK_IMPORT(0:NEIBPETOT))
          ALLOCATE (STACK_EXPORT(0:NEIBPETOT))

          STACK_IMPORT(0)=0
          STACK_EXPORT(0)=0
          NPSI= STACK_IMPORT(NEIBPETOT)
          NPSE= STACK_EXPORT(NEIBPETOT)
          ALLOCATE (NOD_IMPORT(NPSI))
          ALLOCATE (NOD_EXPORT(NPSE))
       ENDIF
    end if


!C reform matrix structure
    DO i = N + 1, NP
       ISTACK_L(i) = ISTACK_L(N)
       ISTACK_U(i) = ISTACK_U(N)
    END DO

!C-- end: matrix input

    allocate(INU(0:N),INL(0:N))
    INU(0:N) = ISTACK_U(0:N)
    INL(0:N) = ISTACK_L(0:N)
    


  CALL MPI_BARRIER(SOLVER_COMM, ierr)
  CALL MPI_BCAST  (METHOD  , 20,       MPI_CHARACTER,               &
       &                   0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (PRECOND  ,20,       MPI_CHARACTER,               &
       &                   0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (REALARRAY(1), 1, MPI_DOUBLE_PRECISION,          &
       &                 0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (REALARRAY(2), 1, MPI_DOUBLE_PRECISION,          &
       &                 0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (REALARRAY(3), 1, MPI_DOUBLE_PRECISION,          &
       &                 0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (INTARRAY(1) , 1, MPI_INTEGER,                   &
       &                 0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (INTARRAY(2) , 1, MPI_INTEGER,                   &
       &                 0, SOLVER_COMM, ierr)
  CALL MPI_BCAST  (dist_f , 1, MPI_LOGICAL,                        &
       &                 0, SOLVER_COMM, ierr)
  CALL MPI_BARRIER(SOLVER_COMM, ierr)

  
  STARTTIME= MPI_WTIME()
  
  !C-- allocation of solution vector
  ALLOCATE( DELPHI (NP) )
  
  DO i = 1, NP
     DELPHI(i) = 0.d0
  END DO

  ITER = 100
  RESID = 1.0E-8
  PRECOND = "SHA2"
  CALL AMGCG    (N, NP,  NPL, NPU, DIAG,  AICCG_L, INL, ITEM_L, AICCG_U, INU, ITEM_U,  &
       & BFORCE, DELPHI, PRECOND, RESID, ITER, ERROR, my_rank, NEIBPETOT, NEIBPE,   &
       & STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
       & PETOT)

  CALL MPI_BARRIER  (SOLVER_COMM,ierr)
  ENDTIME = MPI_WTIME()
  
  IF (my_rank.EQ.0) THEN
     RTIME = ENDTIME-STARTTIME
     WRITE (*, '("*** ELAPSE TIME", 1pe16.6, " sec.")') RTIME
  ENDIF
  
  CALL MPI_FINALIZE(ierr)
END PROGRAM SOLVER_TEST
